# DCRP
# ====

# *D*ynamically *C*onfigurable *R*everse *P*roxy
#
# A http reverse proxy that can have its routing policy reconfigured without
# requiring a restart.
#
# If the "dynamic dispatch" or "late binding" of which backend should handle a
# specific request isn't something that appeals to you then you probably want
# to check out [HAProxy] which is awesome.
#
# [HAProxy]:http://haproxy.1wt.eu
#
# DCRP doesn't have a config file. Clients (backends) must connect to it while
# it's running and provide a list of routes that they are capable of
# responding to.
#
# Clients must also respond to `GET /_ping` in less than a second or DCRP will
# consider them to have gone down and will unload their configuration.
#
# If no client can be found to handle the request then it will be proxied to
# the default client which must be provided to launch DCRP.

# ---

# Modules from node core
http        = require 'http'
url         = require 'url'
dns         = require 'dns'
querystring = require 'querystring'
fs          = require 'fs'

# From NPM
httpProxy = require 'http-proxy'
_         = require 'underscore'

# Cofiguration and Inspection of DCRP is done with
# [Message Sockets](https://github.com/quackingduck/message-socket)
msgSocket = require 'message-socket'
msgSocket.messageFormat = 'json'

msLogger = require 'ms-logger'

# [Message Socket Logger](https://github.com/quackingduck/ms-logger),
# the `null` arg initializes it noop mode
log = msLogger 'das', null

# Public interface, the `dcrp` executable passes the command-line args are
# passed to this function (as an array)
@run = (args) ->
  args = parseCmdlineArgs(args)

  # Default log sockets: `/tmp/mslogger/dcrp` and `/tmp/mslogger/dcrp.pub`
  log.openDefaultSockets()

  log.info "--- starting DAS"

  # Start the HTTP Interface with the one required argument: the default
  # client's url
  httpInterface(args.defaultClientUrl).listen 9000

  # When the program is running, the configs of connected clients are cached
  # on the filesystem so that when the program is killed (or crashes) they can
  # be recovered when it next starts up. We load them here.
  loadClientsFromConfigCache()

  # Start the process that checks to see if connected clients are still alive
  clientHeartbeatMonitor()

  # Open the config message sockets so that new clients can connect and
  # diagnostic tools can monitor the program's state
  configInterface(args.configAddress)

# ### The public facing http server
httpInterface = (defaultClientUrlString) ->
  log.info 'starting http server'

  defaultClientUrl = url.parse defaultClientUrlString
  '' # might be worth resolving-out this hostname
  defaultClient = createProxy
    host: defaultClientUrl.hostname, port: defaultClientUrl.port

  http.createServer (req, res) ->

    # Core proxy logic
    client = clientWithHandlerFor req
    if client
      client.proxyRequest req, res
    else
      defaultClient.proxyRequest req, res

# ### Client Route Matcher
#
# Decides which (if any) client should handle the request. This is the
# computationally expensive part of the program.
clientWithHandlerFor = (req) ->

  # Grab the list of all handlers (route matchers) across all clients
  handlers = registeredHandlers()

  # Early exit optimization
  return if handlers.length is 0

  # Small optimization. No point in doing this more than once.
  req._features = parseFeaturesFromCookies req.headers.cookie
  req._path = url.parse(req.url).pathname

  # Filter to only handlers that match against the request
  handlers = _.select handlers, (h) -> h.matches req

  # Early exit optimization
  return if handlers.length is 0

  # Find the handler with the highest priority
  handler = _.max handlers, (h) -> h.priority()

  # Return the handler's client
  handler.client

# ### Configuration Interface
#
# This is how clients tell DCRP about the routes they handle.
#
# We listen on port 2001 for a message socket pushes of client config json
# structures and on 2002 for diagnostic requests.
#
# DCRP doesn't implement any authentication and a clever attacker might be
# able to crash the process by sending the right kind of garbage over the wire
# so make sure that the config address is on a private network.

configInterface = (address) ->
  address or= '127.0.0.1'
  log.info "starting config sockets on: #{address}"

  configPull = msgSocket.pull "tcp://#{address}:2001"
  configPull (msg) ->
    if msg.msg is 'connect'
      delete msg.msg
      Client.register msg
    else
      log.error "unrecognized message on client config pull socket: #{JSON.stringify msg}"

  configReply = msgSocket.reply "tcp://#{address}:2002"
  configReply (msg, send) ->
    if msg.msg is 'get_config'
      delete msg.msg
      send _.pluck registeredClients(), 'config'
    else
      errorDesc = "unrecognized message on diagnostics reply socket: #{JSON.stringify msg}"
      log.error errorDesc
      send msg: "error", desc: errorDesc


# ### Configuration Update Loop

clientHeartbeatMonitor = ->
  log.info 'starting client heartbeat monitor'
  # How often we check that the clients are up and responding and the minimum
  # amount of time to wait before considering a client to be down. Can't
  # think of a reason why these should ever be a different value.
  pollInterval = clientTimeout = 1000

  setInterval(->

    for client in registeredClients()
      if client.isStillWaitingForPong() and client.msSinceLastPing() >= clientTimeout
        client.deregister()
      else
        client.ping()

  , pollInterval)


# ## Implementation

configCachePath = '/tmp/das-config-cache'
loadClientsFromConfigCache = ->
  fs.readFile configCachePath, (error, data) ->
    if error
      log.info "couldn't read config file, starting with empty config"
    else
      cachedConfig = try JSON.parse data
      if cachedConfig
        log.info "loading last config from config cache"
        Client.register config for config in JSON.parse data
      else
        log.error "couldn't parse config file, starting with empty config"

updateConfigCache = ->
  config = JSON.stringify _.pluck registeredClients(), 'config'
  fs.writeFile configCachePath, config, (error) ->
    if error
      log.error "couldn't write config file"
    else
      log.info "updated config file"


createProxy = (targetAttrs) ->
  new httpProxy.HttpProxy target: targetAttrs

registeredHandlers = ->
  _.flatten _.pluck registeredClients(), 'handlers'

registeredClients = ->
  v for k,v of Client.registeredClients

parseFeaturesFromCookies = (cookieString) ->
  features = querystring.parse(cookieString).features
  features.split ',' if features?

class Client

  # A { nameString -> clientInstance } map of connected clients managed by
  # `Client.register` and `Client#deregister`
  @registeredClients = {}

  # Adds a client
  @register: (config) ->
    client = new Client(config)
    @registeredClients[client.name()] = client
    log.info "registered #{client.name()}"
    updateConfigCache()
    # May be a bit of an aggressive optimization. I was seeing DNS errors
    # on my local machine (when trying to resolve localhost on every request)
    # every so often and doing this completely removed them.
    log.info "resolving dns for #{client.name()} (#{client.config.host})"
    client.isResolvingDNS = yes
    dns.resolve client.config.host, 'A', (err,addresses) ->
      client.ip = _.first addresses
      client.proxy.target.host = client.ip
      client.isResolvingDNS = no
      log.info "resolved dns for #{client.name()} (#{client.config.host})"
      client.ping()

  constructor: (@config) ->
    @handlers = _.map @config.handlers, (h) => new Handler this, h
    # Not sure if this resolves the host on every request
    @proxy = new httpProxy.HttpProxy target: { host: @config.host, port: @config.port }

  # Sets `@pingedAt` to the current timestamp then performs a HTTP GET to
  # `/_ping`. When a successful response is received from the client
  # `@pingedAt` is nulled, if the response is not successful (i.e. not a HTTP
  # 200 OK from the client) for any reason (status not 200, http parse error,
  # dns error, tcp error etc) then `@pingedAt` is left alone and DCRP will reap
  # this client.
  ping: ->
    return if @pingedAt? or @isResolvingDNS
    log.debug "pinging #{@name()} (http://#{@ip}:#{@config.port}/_ping)"
    @pingedAt = new Date
    req = http.get { host: @ip, port: @config.port, path: '/_ping' },
      (res) =>
        if res.statusCode is 200
          log.debug "received pong from #{@name()} in #{(new Date) - @pingedAt}ms"
          @pingedAt = null
        else
          log.warn "got error (#{res.statusCode}) from #{@name()}"
    req.on 'error', (error) =>
      log.error "error pinging #{@name()}"

  proxyRequest: (req, res) ->
    @proxy.proxyRequest req, res

  isStillWaitingForPong: -> @pingedAt?

  msSinceLastPing: -> (new Date) - @pingedAt

  name: -> @config.name

  deregister: () ->
    log.info "deregistering #{@name()}"
    delete @constructor.registeredClients[@name()]
    updateConfigCache()


class Handler

  # This is the hacky part
  @compilePathExp: (pathExp) ->

    # first we convert `/` `**` and `*` into less common chars
    pathExp = pathExp.replace /\//g,   '∕' # ∕ = DIVISON SLASH (2215)
    pathExp = pathExp.replace /\*\*/g, '‥' # ‥ = TWO DOT LEADER (2025)
    pathExp = pathExp.replace /\*/g,   '․' # ․ = ONE DOT LEADER (2024)

    # then we replace '․' and '‥' with appropriate regex patterns
    pathExp = pathExp.replace(/․/g,'[^∕]+') # single segment
    pathExp = pathExp.replace(/‥/g,'[^.]+') # multi segment

    # then we add the anchors
    pathExp = "^#{pathExp}$"
    new RegExp pathExp

  @unescapeRequestPath: (rawRequestPath) ->
    # To prepare a raw path for comparison with the pathExp it needs to be
    # unscaped and have it's `/` segment separator replaced with `∕`
    (querystring.unescape(seg) for seg in rawRequestPath.split '/').join '∕'

  constructor: (@client,@config) ->
    @pathRegex = @constructor.compilePathExp @config.path

  matches: (request) ->
    @supportsFeatureIn(request._features) and @pathMatches(request._path)

  # Feature detection:
  supportsFeatureIn: (features) ->
    # this handler only applies to requests with a specific feature enabled
    if @config.feature?
      # this request has no features enabled so it can't match
      return no unless features?
      # try to find this feature in the enabled ones for this request
      _.include features, @config.feature
    # this handler applies to all features
    else yes

  # Path matching:
  pathMatches: (requestPath) ->
    @pathRegex.test @constructor.unescapeRequestPath(requestPath)

  path: -> @config.path
  feature: -> @config.feature

  priority: ->
    @config.priority or 1

parseCmdlineArgs = (rawArgs) ->
  parsedArgs = {}
  while arg = rawArgs.shift()
    switch arg
      when '--help', '-h', '-?' then usage()
      when '--config-address' then parsedArgs.configAddress = rawArgs.shift()
      else parsedArgs.defaultClientUrl = arg if rawArgs.length is 0
  return usage() unless parsedArgs.defaultClientUrl?
  parsedArgs

usage = ->
  console.log """
  Usage:
    dcrp [options] <default-client-url>

  Required:
    <default-client-url>
      The url to proxy all requests that aren't handled by a connected client

  Options:
    --config-address <ip-address>
      The IP to listen for config messages on, by default this is the loopback
      interface (127.0.0.1).

      Client config push messages will be received on port 2001
      System diagnostics can be retreived from port 2002

    -h|--help|-?
      Display these usage instructions.

  Examples:
    dcrp http://localhost:81
    dcrp --config-address 10.1.3.29 http://localhost:81
  """
  process.exit()

# Expose some internals for testing
@_test = { Handler }
