(function() {
  var Client, Handler, clientHeartbeatMonitor, clientWithHandlerFor, configCachePath, configInterface, createProxy, dns, fs, http, httpInterface, httpProxy, loadClientsFromConfigCache, log, msLogger, msgSocket, parseCmdlineArgs, parseFeaturesFromCookies, querystring, registeredClients, registeredHandlers, updateConfigCache, url, usage, _;
  var __bind = function(fn, me){ return function(){ return fn.apply(me, arguments); }; };
  http = require('http');
  url = require('url');
  dns = require('dns');
  querystring = require('querystring');
  fs = require('fs');
  httpProxy = require('http-proxy');
  _ = require('underscore');
  msgSocket = require('message-socket');
  msgSocket.messageFormat = 'json';
  msLogger = require('ms-logger');
  log = msLogger('das', null);
  this.run = function(args) {
    args = parseCmdlineArgs(args);
    log.openDefaultSockets();
    log.info("--- starting DAS");
    httpInterface(args.defaultClientUrl).listen(9000);
    loadClientsFromConfigCache();
    clientHeartbeatMonitor();
    return configInterface(args.configAddress);
  };
  httpInterface = function(defaultClientUrlString) {
    var defaultClient, defaultClientUrl;
    log.info('starting http server');
    defaultClientUrl = url.parse(defaultClientUrlString);
    '';
    defaultClient = createProxy({
      host: defaultClientUrl.hostname,
      port: defaultClientUrl.port
    });
    return http.createServer(function(req, res) {
      var client;
      client = clientWithHandlerFor(req);
      if (client) {
        return client.proxyRequest(req, res);
      } else {
        return defaultClient.proxyRequest(req, res);
      }
    });
  };
  clientWithHandlerFor = function(req) {
    var handler, handlers;
    handlers = registeredHandlers();
    if (handlers.length === 0) {
      return;
    }
    req._features = parseFeaturesFromCookies(req.headers.cookie);
    req._path = url.parse(req.url).pathname;
    handlers = _.select(handlers, function(h) {
      return h.matches(req);
    });
    if (handlers.length === 0) {
      return;
    }
    handler = _.max(handlers, function(h) {
      return h.priority();
    });
    return handler.client;
  };
  configInterface = function(address) {
    var configPull, configReply;
    address || (address = '127.0.0.1');
    log.info("starting config sockets on: " + address);
    configPull = msgSocket.pull("tcp://" + address + ":2001");
    configPull(function(msg) {
      if (msg.msg === 'connect') {
        delete msg.msg;
        return Client.register(msg);
      } else {
        return log.error("unrecognized message on client config pull socket: " + (JSON.stringify(msg)));
      }
    });
    configReply = msgSocket.reply("tcp://" + address + ":2002");
    return configReply(function(msg, send) {
      var errorDesc;
      if (msg.msg === 'get_config') {
        delete msg.msg;
        return send(_.pluck(registeredClients(), 'config'));
      } else {
        errorDesc = "unrecognized message on diagnostics reply socket: " + (JSON.stringify(msg));
        log.error(errorDesc);
        return send({
          msg: "error",
          desc: errorDesc
        });
      }
    });
  };
  clientHeartbeatMonitor = function() {
    var clientTimeout, pollInterval;
    log.info('starting client heartbeat monitor');
    pollInterval = clientTimeout = 1000;
    return setInterval(function() {
      var client, _i, _len, _ref, _results;
      _ref = registeredClients();
      _results = [];
      for (_i = 0, _len = _ref.length; _i < _len; _i++) {
        client = _ref[_i];
        _results.push(client.isStillWaitingForPong() && client.msSinceLastPing() >= clientTimeout ? client.deregister() : client.ping());
      }
      return _results;
    }, pollInterval);
  };
  configCachePath = '/tmp/das-config-cache';
  loadClientsFromConfigCache = function() {
    return fs.readFile(configCachePath, function(error, data) {
      var cachedConfig, config, _i, _len, _ref, _results;
      if (error) {
        return log.info("couldn't read config file, starting with empty config");
      } else {
        cachedConfig = (function() {
          try {
            return JSON.parse(data);
          } catch (_e) {}
        })();
        if (cachedConfig) {
          log.info("loading last config from config cache");
          _ref = JSON.parse(data);
          _results = [];
          for (_i = 0, _len = _ref.length; _i < _len; _i++) {
            config = _ref[_i];
            _results.push(Client.register(config));
          }
          return _results;
        } else {
          return log.error("couldn't parse config file, starting with empty config");
        }
      }
    });
  };
  updateConfigCache = function() {
    var config;
    config = JSON.stringify(_.pluck(registeredClients(), 'config'));
    return fs.writeFile(configCachePath, config, function(error) {
      if (error) {
        return log.error("couldn't write config file");
      } else {
        return log.info("updated config file");
      }
    });
  };
  createProxy = function(targetAttrs) {
    return new httpProxy.HttpProxy({
      target: targetAttrs
    });
  };
  registeredHandlers = function() {
    return _.flatten(_.pluck(registeredClients(), 'handlers'));
  };
  registeredClients = function() {
    var k, v, _ref, _results;
    _ref = Client.registeredClients;
    _results = [];
    for (k in _ref) {
      v = _ref[k];
      _results.push(v);
    }
    return _results;
  };
  parseFeaturesFromCookies = function(cookieString) {
    var features;
    features = querystring.parse(cookieString).features;
    if (features != null) {
      return features.split(',');
    }
  };
  Client = (function() {
    Client.registeredClients = {};
    Client.register = function(config) {
      var client;
      client = new Client(config);
      this.registeredClients[client.name()] = client;
      log.info("registered " + (client.name()));
      updateConfigCache();
      log.info("resolving dns for " + (client.name()) + " (" + client.config.host + ")");
      client.isResolvingDNS = true;
      return dns.resolve(client.config.host, 'A', function(err, addresses) {
        client.ip = _.first(addresses);
        client.proxy.target.host = client.ip;
        client.isResolvingDNS = false;
        log.info("resolved dns for " + (client.name()) + " (" + client.config.host + ")");
        return client.ping();
      });
    };
    function Client(config) {
      this.config = config;
      this.handlers = _.map(this.config.handlers, __bind(function(h) {
        return new Handler(this, h);
      }, this));
      this.proxy = new httpProxy.HttpProxy({
        target: {
          host: this.config.host,
          port: this.config.port
        }
      });
    }
    Client.prototype.ping = function() {
      var req;
      if ((this.pingedAt != null) || this.isResolvingDNS) {
        return;
      }
      log.debug("pinging " + (this.name()) + " (http://" + this.ip + ":" + this.config.port + "/_ping)");
      this.pingedAt = new Date;
      req = http.get({
        host: this.ip,
        port: this.config.port,
        path: '/_ping'
      }, __bind(function(res) {
        if (res.statusCode === 200) {
          log.debug("received pong from " + (this.name()) + " in " + ((new Date) - this.pingedAt) + "ms");
          return this.pingedAt = null;
        } else {
          return log.warn("got error (" + res.statusCode + ") from " + (this.name()));
        }
      }, this));
      return req.on('error', __bind(function(error) {
        return log.error("error pinging " + (this.name()));
      }, this));
    };
    Client.prototype.proxyRequest = function(req, res) {
      return this.proxy.proxyRequest(req, res);
    };
    Client.prototype.isStillWaitingForPong = function() {
      return this.pingedAt != null;
    };
    Client.prototype.msSinceLastPing = function() {
      return (new Date) - this.pingedAt;
    };
    Client.prototype.name = function() {
      return this.config.name;
    };
    Client.prototype.deregister = function() {
      log.info("deregistering " + (this.name()));
      delete this.constructor.registeredClients[this.name()];
      return updateConfigCache();
    };
    return Client;
  })();
  Handler = (function() {
    Handler.compilePathExp = function(pathExp) {
      pathExp = pathExp.replace(/\//g, '∕');
      pathExp = pathExp.replace(/\*\*/g, '‥');
      pathExp = pathExp.replace(/\*/g, '․');
      pathExp = pathExp.replace(/․/g, '[^∕]+');
      pathExp = pathExp.replace(/‥/g, '[^.]+');
      pathExp = "^" + pathExp + "$";
      return new RegExp(pathExp);
    };
    Handler.unescapeRequestPath = function(rawRequestPath) {
      var seg;
      return ((function() {
        var _i, _len, _ref, _results;
        _ref = rawRequestPath.split('/');
        _results = [];
        for (_i = 0, _len = _ref.length; _i < _len; _i++) {
          seg = _ref[_i];
          _results.push(querystring.unescape(seg));
        }
        return _results;
      })()).join('∕');
    };
    function Handler(client, config) {
      this.client = client;
      this.config = config;
      this.pathRegex = this.constructor.compilePathExp(this.config.path);
    }
    Handler.prototype.matches = function(request) {
      return this.supportsFeatureIn(request._features) && this.pathMatches(request._path);
    };
    Handler.prototype.supportsFeatureIn = function(features) {
      if (this.config.feature != null) {
        if (features == null) {
          return false;
        }
        return _.include(features, this.config.feature);
      } else {
        return true;
      }
    };
    Handler.prototype.pathMatches = function(requestPath) {
      return this.pathRegex.test(this.constructor.unescapeRequestPath(requestPath));
    };
    Handler.prototype.path = function() {
      return this.config.path;
    };
    Handler.prototype.feature = function() {
      return this.config.feature;
    };
    Handler.prototype.priority = function() {
      return this.config.priority || 1;
    };
    return Handler;
  })();
  parseCmdlineArgs = function(rawArgs) {
    var arg, parsedArgs;
    parsedArgs = {};
    while (arg = rawArgs.shift()) {
      switch (arg) {
        case '--help':
        case '-h':
        case '-?':
          usage();
          break;
        case '--config-address':
          parsedArgs.configAddress = rawArgs.shift();
          break;
        default:
          if (rawArgs.length === 0) {
            parsedArgs.defaultClientUrl = arg;
          }
      }
    }
    if (parsedArgs.defaultClientUrl == null) {
      return usage();
    }
    return parsedArgs;
  };
  usage = function() {
    console.log("Usage:\n  dcrp [options] <default-client-url>\n\nRequired:\n  <default-client-url>\n    The url to proxy all requests that aren't handled by a connected client\n\nOptions:\n  --config-address <ip-address>\n    The IP to listen for config messages on, by default this is the loopback\n    interface (127.0.0.1).\n\n    Client config push messages will be received on port 2001\n    System diagnostics can be retreived from port 2002\n\n  -h|--help|-?\n    Display these usage instructions.\n\nExamples:\n  dcrp http://localhost:81\n  dcrp --config-address 10.1.3.29 http://localhost:81");
    return process.exit();
  };
  this._test = {
    Handler: Handler
  };
}).call(this);
